#!/usr/bin/env python
# Coolan Confidential
# Copyright (c) Coolan, Inc 2014 All Rights Reserved
#
# NOTICE:  All information contained herein is, and remains
# the property of Coolan, Inc.  The intellectual and
# technical concepts contained #herein are proprietary to
# Coolan, Inc and may be covered by U.S. and Foreign Patents,
# patents in process, and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material
# is strictly forbidden unless prior written permission is obtained
# from Coolan, Inc.
# This code and information are provided "as is" without warranty of any
# kind, either expressed or implied, including but not limited to the
# implied warranties of mechantability and/or fitness for a
# particular purpose.

import os
import re
import subprocess  # make system calls.
import sys
import shlex
import ConfigParser
import threading
import uuid
import base64
import warnings
import time
from datetime import datetime
import errno
import fileinput
import traceback

import settings as s


FAIL = '\033[91m'
ENDC = '\033[0m'

script_path = os.path.dirname(os.path.realpath(sys.argv[0]))

sys.path.append(os.path.join(script_path, 'modules'))
import requests


def read_config(file_path_list):
    """
    Check if the properties file exists and if the format is correct.
    """
    local_config = ConfigParser.SafeConfigParser()
    # preserve the case of the option keys
    local_config.optionxform = str
    for file_path in file_path_list:
        if not os.path.isfile(file_path):
            print "\nERROR: %s does not exist\n" % file_path
            sys.exit(1)
    local_config.read(file_path_list)
    return local_config


def get_config_file():
    return os.path.join(script_path, s.CONFIG_FILE)


def get_default_config_file():
    return os.path.join(script_path, s.DEFAULT_CONFIG_FILE)


config = read_config([get_default_config_file(), get_config_file()])

os.nice(int(config.get(s.SETTINGS, "python_nice_level")))


def get_split_pattern():
    delimiters = s.DELIMITERS
    return re.compile(delimiters)


def replace(file_path, pattern, subst):
    # Read contents from file as a single string
    file_handle = open(file_path, 'r')
    file_string = file_handle.read()
    file_handle.close()

    # Use RE package to allow for replacement (also allowing for (multiline) REGEX)
    file_string = (re.sub(pattern, subst, file_string))

    # Write contents to file.
    # Using mode 'w' truncates the file.
    file_handle = open(file_path, 'w')
    file_handle.write(file_string)
    file_handle.close()


def set_commands_path():
    env_copy = os.environ.copy()
    os_path = os.getenv('PATH').split(os.pathsep)

    try:
        split_pattern = get_split_pattern()
        commands_path = split_pattern.split(config.get(s.FILE_LOCATIONS, 'commands_path').strip())
        if not set(os_path).issubset(set(commands_path)):
            merged_path = list(set(commands_path + os_path))
            env_copy['PATH'] = ':'.join(merged_path)
            pattern = "commands_path=.+"
            new_config_string = "commands_path=" + env_copy['PATH']
            replace(get_config_file(), pattern, new_config_string)
            config.set(s.FILE_LOCATIONS, 'commands_path', env_copy['PATH'])
        else:
            env_copy['PATH'] = ':'.join(set(commands_path))
        return env_copy
    except ConfigParser.NoOptionError:
        pass


current_env = set_commands_path()


def print_fail(msg):
    print FAIL + msg + ENDC


def verify_root():
    """
    Verify that the script is being run as root, bail otherwise.
    """
    if not os.geteuid() == 0:
        print("\nOnly root can run this script\n")
        sys.exit(1)


def get_timestamp():
    return datetime.now().strftime("%Y-%m-%d %H:%M:%S")


def run_subprocess_popen(args, via_shell=False):
    if via_shell:
        p = subprocess.Popen(" ".join(args), stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True, close_fds=True)
    else:
        p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
    return p


def get_command_output(args, via_shell=False, add_nice_wrapper=True):
    timeout = int(config.get(s.SETTINGS, 'periodic_commands_time_out_seconds'))
    start = time.time()
    output_string = ""
    err_string = ""
    if add_nice_wrapper:
        # adding the nice wrapper to the command
        args = shlex.split(global_nice_wrapper) + args

    p = run_subprocess_popen(args, via_shell)
    while p.poll() is None:
        for line in iter(p.stdout.readline, ''):
            output_string += line
            if time.time() - start > timeout:
                terminate_subprocess(p, args)
                err_string += s.COMMAND_TIMEOUT_MSG % (get_timestamp(), " ".join(args), config.get(s.SETTINGS, 'periodic_commands_time_out_seconds'))
                return "", err_string

        for err_line in iter(p.stderr.readline, ''):
            err_string += err_line
            if time.time() - start > timeout:
                terminate_subprocess(p, args)
                err_string += s.COMMAND_TIMEOUT_MSG % (get_timestamp(), " ".join(args), config.get(s.SETTINGS, 'periodic_commands_time_out_seconds'))
                return "", err_string
    out, err = p.communicate()
    output_string += out
    err_string += err
    return_code = p.returncode
    if return_code != 0:
        err_string += "\nReturn Code: " + str(return_code)
    return output_string, err_string


def find_command(cmd, full_cmd=""):
    if cmd in set(s.XENTOOL) or cmd in set(s.HPTOOL):
        command_full_path = find_tool(cmd, full_cmd)
    else:
        args = shlex.split("which " + cmd)
        command_full_path, command_full_path_err = get_command_output(args=args, add_nice_wrapper=False)

        if cmd in set(s.MEGACLI):
            if not command_full_path.strip():
                command_full_path = find_megacli_command(cmd)

    return command_full_path.strip()


def generate_nice_wrapper():
    """
    Generate the nice command to lower the priority of the running command to minimize the impact of the system
    """
    nice_full_path = find_command("nice")
    ionice_full_path = find_command("ionice")
    nocache_full_path = find_command("nocache")
    nice_wrapper = ""
    if nice_full_path:
        nice_wrapper = nice_full_path + " -n -" + config.get(s.SETTINGS, "commands_nice_level")
    if ionice_full_path:
        nice_wrapper = ionice_full_path + " -c" + config.get(s.SETTINGS, "commands_ionice_class") + " " + nice_wrapper
    if nocache_full_path:
        nice_wrapper = nocache_full_path + " " + nice_wrapper

    return nice_wrapper


global_nice_wrapper = generate_nice_wrapper()


def create_directory(folder_path):
    if not os.path.exists(folder_path):
        os.makedirs(folder_path)


def get_new_machine_id():
    """
    D-Bus UUID remains unchanged until the next reboot and it is different for any two running instances of the OS kernel.
    must be exactly 32 characters long.
    """
    machine_id = ""
    try:
        split_pattern = get_split_pattern()
        machine_id_file_locations = split_pattern.split(config.get(s.FILE_LOCATIONS, 'machine_id_file').strip())
        # Search for the machine id in the provided locations
        for machine_id_file_location in machine_id_file_locations:
            if os.path.isfile(machine_id_file_location.strip()):
                out = run_cat_command("cat " + machine_id_file_location.strip())
                if out.strip():
                    machine_id = out.strip()

        # If machine Id is not found on the machine ,get the MACHINE_ID in the setting file.(So it doesn't generate random machine Id every time)
        if not machine_id and hasattr(s, 'MACHINE_ID') and s.MACHINE_ID.strip():
            machine_id = s.MACHINE_ID.strip()
            print_fail("ERROR: Machine Id not found in any of the following locations:\n%s\nRandom Machine Id %s will be used.\n" % ('\n'.join(machine_id_file_locations), machine_id))
        # Generate one if there's no machine Id in the setting file
        elif not machine_id:
            machine_id = "random" + uuid.uuid4().hex[6:]
            print_fail("ERROR: Machine Id not found in any of the following locations:\n%s\nRandom Machine Id %s will be used.\n" % ('\n'.join(machine_id_file_locations), machine_id))
        add_property_to_settings('MACHINE_ID', machine_id)
        return machine_id
    # if 'machine_id_file' is not in config file
    except ConfigParser.NoOptionError:
        if not get_machine_id_setting():
            machine_id = "random" + uuid.uuid4().hex[6:]
            print_fail("ERROR: machine_id_file not found in %s. Random Machine Id %s will be used." % (get_config_file(), machine_id))
        else:
            machine_id = s.MACHINE_ID
        add_property_to_settings('MACHINE_ID', machine_id)
        return machine_id


def get_machine_id_setting():
    if hasattr(s, 'MACHINE_ID'):
        return s.MACHINE_ID.strip()
    else:
        return ""


def get_new_boot_id():
    boot_id = ""
    try:
        split_pattern = get_split_pattern()
        boot_id_file_locations = split_pattern.split(config.get(s.FILE_LOCATIONS, 'boot_id_file').strip())
        # Search for the boot id in the provided locations
        for boot_id_file_location in boot_id_file_locations:
            if os.path.isfile(boot_id_file_location.strip()):
                out = run_cat_command("cat " + boot_id_file_location.strip())
                if out.strip():
                    boot_id = out.strip()

        # If boot Id is not found on the machine ,get the BOOT_ID in the setting file.(So it doesn't generate random boot Id every time)
        if not boot_id and hasattr(s, 'BOOT_ID') and s.BOOT_ID.strip():
            boot_id = s.BOOT_ID.strip()
            print_fail("ERROR: Boot Id not found in any of the following locations:\n%s\nRandom Boot Id %s will be used.\n" % ('\n'.join(boot_id_file_locations), boot_id))
        # Generate one if there's no boot Id in the setting file
        elif not boot_id:
            boot_id = "random" + str(uuid.uuid4())[6:]
            print_fail("ERROR: Boot Id not found in any of the following locations:\n%s\nRandom Boot Id %s will be used.\n" % ('\n'.join(boot_id_file_locations), boot_id))
        add_property_to_settings('BOOT_ID', boot_id)
        return boot_id
    # if 'boot_id_file' is not in config file
    except ConfigParser.NoOptionError:
        if not get_boot_id_setting():
            boot_id = "random" + str(uuid.uuid4())[6:]
            print_fail("ERROR: 'boot_id_file' not found in %s. Random Boot Id %s will be used." % (get_config_file(), boot_id))
        else:
            boot_id = s.BOOT_ID
        add_property_to_settings('BOOT_ID', boot_id)
        return boot_id


def get_boot_id_setting():
    if hasattr(s, 'BOOT_ID'):
        return s.BOOT_ID.strip()
    else:
        return ""


def add_property_to_settings(property_name, value):
    if property_name and value:
        # add the setting if the setting does not exist
        if not hasattr(s, property_name):
            settings_file = open(get_settings_file(), 'a')
            settings_file.write('\n' + property_name + ' = ' + str(value))
            settings_file.close()
        # re-write the setting value if the setting already exist
        elif value != getattr(s, property_name):
            pattern = property_name + " =.+"
            new_setting_string = property_name + " = '" + str(value) + "'"
            replace(get_settings_file(), pattern, new_setting_string)
        setattr(s, property_name, value)


def get_last_run_daily_time():
    """
    get the last time the daily commands were run
    """
    if hasattr(s, 'LAST_RUN_DAILY'):
        return float(s.LAST_RUN_DAILY)
    else:
        return 0


def get_last_run_time():
    """
    get the last time the collector was run not in test mode
    """
    if hasattr(s, 'LAST_RUN'):
        return float(s.LAST_RUN)
    else:
        return 0


def get_customer_id():
    try:
        customer_id = config.get(s.AUTH_SECTION, 'customerID').strip()
        if customer_id:
            return customer_id
        else:
            return "unknown"
    except ConfigParser.NoOptionError:
        return "unknown"


def get_username():
    try:
        return config.get(s.AUTH_SECTION, 'username').strip()
    except ConfigParser.NoOptionError:
        return ""


def get_password():
    try:
        return config.get(s.AUTH_SECTION, 'password').strip()
    except ConfigParser.NoOptionError:
        return ""


def get_endpoint():
    try:
        return config.get(s.AUTH_SECTION, 'logs_endpoint').strip()
    except ConfigParser.NoOptionError:
        return ""


def verify_auth_section():
    if get_customer_id() == "unknown":
        print_fail("ERROR: Customer Id not found in %s" % get_config_file())

    if not get_username():
        print_fail("ERROR: Username not found in %s" % get_config_file())

    if not get_password():
        print_fail("ERROR: Password not found in %s" % get_config_file())


def get_proxy_url():
    try:
        return config.get(s.AUTH_SECTION, 'proxyUrl').strip()
    except ConfigParser.NoOptionError:
        return ""


def get_commands_file():
    return os.path.join(script_path, s.COMMANDS_FILE)


commands_config = read_config([get_commands_file()])


def get_commands(section):
    """
    Get all commands for the given section
    """
    return set(commands_config.options(section))


def get_all_commands_and_description():
    """
    Get all key-value properties for the given section and return a map
    """
    commands_map = {}
    for section in commands_config.sections():
        options = commands_config.options(section)
        for option in options:
            commands_map[option] = commands_config.get(section, option)
    return commands_map


def generate_log_filename():
    return s.RAWDATA+"-%sM%sB%sT%s" % (get_customer_id(), get_machine_id_setting(), get_boot_id_setting(), int(time.time()))


def generate_verify_log_filename():
    return "verify-%sM%sB%sT%s.log" % (get_customer_id(), get_machine_id_setting(), get_boot_id_setting(), int(time.time()))


def generate_time_series_log_filename():
    return s.TIME_SERIES+"-%sM%sB%sT%s" % (get_customer_id(), get_machine_id_setting(), get_boot_id_setting(), int(time.time()))


def generate_time_series_individual_log_filename(command):
    return "individual-timeseries-%s-%sM%sB%sT%s.log%s" % (command.replace(" ", ""), get_customer_id(), get_machine_id_setting(), get_boot_id_setting(), int(time.time()), s.TMP_FILE_EXTENSION)


def generate_custom_log_filename():
    return s.CUSTOM_DATA+"-%sM%sB%sT%s" % (get_customer_id(), get_machine_id_setting(), get_boot_id_setting(), int(time.time()))


def find_megacli_command(cmd):
    """
    This method will only be call when the given megacli commands are not found on the system
    """
    candidate_megacli_commands = s.MEGACLI[:]
    # only looking for substitute megacli command
    candidate_megacli_commands.remove(cmd)
    for candidate_megacli_command in candidate_megacli_commands:
        args = shlex.split("which " + candidate_megacli_command.strip())
        command_full_path, command_full_path_err = get_command_output(args=args, add_nice_wrapper=False)
        if command_full_path.strip():
            return command_full_path.strip()
    return ""


def find_tool(cmd, full_cmd):
    args = shlex.split("which " + cmd)
    command_full_path, command_full_path_err = get_command_output(args=args, add_nice_wrapper=False)
    if command_full_path.strip():
        args = []
        args.extend(shlex.split(full_cmd))
        args[0] = command_full_path.strip()
        stdout, stderr = get_command_output(args=args)
        if stderr.strip() or has_errors_in_stdout(stdout):
            command_full_path = find_alternate_tool(cmd, command_full_path, full_cmd)
    else:
        command_full_path = find_alternate_tool(cmd, command_full_path, full_cmd)
    return command_full_path


def find_alternate_tool(cmd, command_full_path, exec_cmd):
    """
    This method will only be call when the given tool is not found on the system
    """
    if cmd in set(s.XENTOOL):
        alternate_tools = s.XENTOOL[:]
    elif cmd in set(s.HPTOOL):
        alternate_tools = s.HPTOOL[:]
    else:
        alternate_tools = []

    # only looking for substitute tools
    alternate_tools.remove(cmd)
    for alternate_tool in alternate_tools:
        args = shlex.split("which " + alternate_tool.strip())
        command_full_path, command_full_path_err = get_command_output(args=args, add_nice_wrapper=False)
        if command_full_path.strip():
            args = shlex.split(command_full_path)
            args.extend(shlex.split(exec_cmd)[1:])
            stdout, stderr = get_command_output(args=args)
            if not stderr.strip() and not has_errors_in_stdout(stdout):
                return command_full_path.strip()
    return""


def has_errors_in_stdout(stdout):
    for error in s.ERROR_OUTPUT:
        if stdout.strip().lower().startswith(error.lower()):
            return True
    return False


def run_cat_command(cmd):
    args = []
    args.extend(shlex.split(cmd))
    command_full_path = find_command(args[0])
    if command_full_path:
        args[0] = command_full_path
        out, err = get_command_output(args=args)
        return out
    else:
        sys.stderr.write("ERROR: Command %s doesn't exist.\n" % args[0])
        return ''


def post_file(url, username, password, zipPath):
    is_post_success = False
    base64string = base64.encodestring('%s:%s' % (username, password)).replace('\n', '')

    proxy = {'https': get_proxy_url()}

    try:
        warnings.filterwarnings('ignore', '.*Unverified HTTPS*', )
        files = {'file': open(zipPath)}  # 'file' => name of html input field
        auth_header = {'Authorization': 'Basic %s' % base64string}
        response = requests.post(url, verify=False, files=files, headers=auth_header, proxies=proxy)
        if response.status_code == requests.codes.created:
            is_post_success = True
        else:
            print_fail("%s : ERROR: Posting log files to %s failed. Response Code: %s %s" % (get_timestamp(), url, response.status_code, response.reason))
            if response.status_code == requests.codes.unauthorized:
                print_fail('%s : Unable to make connections to Coolan endpoint. Please verify your "customerID", "username", and "password" in %s' % (get_timestamp(), get_config_file()))
    except ValueError:
        if proxy:
            print_fail("%s : ERROR: Posting log files to %s failed. Using proxy: %s" % (get_timestamp(), url, proxy))
        else:
            print_fail("%s : ERROR: Posting log files to %s failed." % (get_timestamp(), url))
    except Exception as e:
        print_fail("%s : ERROR: Posting log files to %s failed. %s" % (get_timestamp(), url, e))
    return is_post_success


def get_settings_file():
    return os.path.join(script_path, 'settings.py')


def get_logs_folder_path():
    try:
        logs_folder_path = os.path.join('/', config.get(s.DIRECTORIES, 'abs_log_file_dir').strip())
    except ConfigParser.NoOptionError:
        print("\nERROR: config 'abs_log_file_dir' does not exist.\n")
        sys.exit(1)

    create_directory(logs_folder_path)
    return os.path.normpath(logs_folder_path)


def get_backup_folder_path():
    try:
        backup_folder_path = os.path.join('/', config.get(s.DIRECTORIES, 'abs_backup_log_folder').strip())
    except ConfigParser.NoOptionError:
        print("\nERROR: config 'abs_backup_log_folder' does not exist.\n")
        sys.exit(1)

    create_directory(backup_folder_path)
    return os.path.normpath(backup_folder_path)


def get_custom_logs_folder_path():
    try:
        custom_logs_folder_path = os.path.join('/', config.get(s.DIRECTORIES, 'abs_custom_log_file_dir').strip())
    except ConfigParser.NoOptionError:
        print("\nERROR: config 'abs_custom_log_file_dir' does not exist.\n")
        sys.exit(1)

    create_directory(custom_logs_folder_path)
    return os.path.normpath(custom_logs_folder_path)


def get_custom_backup_folder_path():
    try:
        custom_backup_folder_path = os.path.join('/', config.get(s.DIRECTORIES, 'abs_custom_backup_log_folder').strip())
    except ConfigParser.NoOptionError:
        print("\nERROR: config 'abs_custom_backup_log_folder' does not exist.\n")
        sys.exit(1)

    create_directory(custom_backup_folder_path)
    return os.path.normpath(custom_backup_folder_path)


def is_pid_running(pid):
    """Check whether pid exists in the current process table.
    UNIX only.
    """
    if pid < 0:
        return False
    if pid == 0:
        # According to "man 2 kill" PID 0 refers to every process
        # in the process group of the calling process.
        # On certain systems 0 is a valid PID but we have no way
        # to know that in a portable fashion.
        raise ValueError('invalid PID 0')
    try:
        os.kill(pid, 0)
    except OSError as err:
        if err.errno == errno.ESRCH:
            # ESRCH == No such process
            return False
        elif err.errno == errno.EPERM:
            # EPERM clearly means there's a process to deny access to
            return True
        else:
            # According to "man 2 kill" possible error values are
            # (EINVAL, EPERM, ESRCH)
            raise
    else:
        return True


class RunCommandInThread(object):
    def __init__(self, cmd, log_path, interval):
        self.cmd = cmd
        self.process = None
        self.log_path = log_path
        self.interval = interval

    def run(self, timeout):
        def target():
            start_time = time.time()
            self.process = subprocess.Popen(self.cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = self.process.communicate()
            write_timeseries_output_to_log(stdout, stderr, self.cmd, self.log_path, self.interval, start_time)

        thread = threading.Thread(target=target)
        thread.start()

        thread.join(timeout)
        if thread.is_alive():
            print "%s : Collecting %s data is taking too long. Killing the command. " % (get_timestamp(), " ".join(self.cmd))
            self.process.terminate()
            thread.join()


def write_command_output_to_log(out, err, args, log_path):
    begin_msg = s.BEGIN_MSG % (" ".join(args))
    end_msg = s.END_MSG % (" ".join(args))
    begin_err_msg = s.BEGIN_ERROR_MSG % (" ".join(args))
    end_err_msg = s.END_ERROR_MSG % (" ".join(args))
    log_file = open(log_path, 'a')
    if out.strip():
        log_file.write(begin_msg + out + end_msg)
    if err.strip():
        log_file.write(begin_err_msg + err + end_err_msg)
    log_file.close()


def write_timeseries_output_to_log(out, err, args, log_path, interval, start_time):
    begin_msg = s.BEGIN_TIMESERIES_MSG % (int(start_time), interval, " ".join(args))
    end_msg = s.END_TIMESERIES_MSG % (int(start_time), interval, " ".join(args))
    begin_err_msg = s.BEGIN_ERROR_TIMESERIES_MSG % (int(start_time), interval, " ".join(args))
    end_err_msg = s.END_ERROR_TIMESERIES_MSG % (int(start_time), interval, " ".join(args))
    log_file = open(log_path, 'a')
    if out.strip():
        log_file.write(begin_msg + out + end_msg)
    if err.strip():
        log_file.write(begin_err_msg + err + end_err_msg)
    log_file.close()


def terminate_subprocess(process, args):
    process.terminate()
    print s.COMMAND_TIMEOUT_MSG % (get_timestamp(), " ".join(args), config.get(s.SETTINGS, 'periodic_commands_time_out_seconds'))


def is_config_on(config_name):
    try:
        if str(config.get(s.SETTINGS, config_name)).lower() in ['y', 'yes', 'true', 't']:
            return True
        else:
            return False
    except ConfigParser.NoOptionError:
        return False


def is_input_string_true(input_string):
    if input_string.lower() in ['y', 'yes', 'true', 't']:
        return True
    else:
        return False


def print_dmidecode_info(log_file):
    dmidecode_map = {"-s system-serial-number": "systemSerialNumber", "-s system-uuid": "systemUuid",
                     "-s baseboard-serial-number": "baseboardSerialNumber",
                     "-s chassis-serial-number": "chassisSerialNumber"}
    dmidecode_cmd = "dmidecode"
    dmidecode_full_path = find_command(dmidecode_cmd)
    if dmidecode_full_path:
        for key, value in dmidecode_map.items():
            dmidecode_args = [dmidecode_full_path]
            dmidecode_args.extend(shlex.split(key))
            out, err = get_command_output(args=dmidecode_args, add_nice_wrapper=True)
            log_file.write(value + " : %s\n" % out.strip())