#!/usr/bin/env python
# Coolan Confidential
# Copyright (c) Coolan, Inc 2014 All Rights Reserved
#
# NOTICE:  All information contained herein is, and remains
# the property of Coolan, Inc.  The intellectual and
# technical concepts contained #herein are proprietary to
# Coolan, Inc and may be covered by U.S. and Foreign Patents,
# patents in process, and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material
# is strictly forbidden unless prior written permission is obtained
# from Coolan, Inc.
# This code and information are provided "as is" without warranty of any
# kind, either expressed or implied, including but not limited to the
# implied warranties of mechantability and/or fitness for a
# particular purpose.

import zipfile
import shutil
import glob

from common import *


def zip_log_files(file_paths, zipfile_name):
    """
    put the given log files in a zip file
    return the filename of the generated zip file
    """
    zipfile_path = os.path.join(get_logs_folder_path(), zipfile_name)
    zf = zipfile.ZipFile(zipfile_path, 'w', zipfile.ZIP_DEFLATED)

    for file_path in file_paths:
        zf.write(file_path, os.path.basename(file_path))
    zf.close()
    print "%s : %s log files were added into zip file %s ." % (get_timestamp(), len(file_paths), zipfile_path)
    # remove the original log files
    for file_path in file_paths:
        os.remove(file_path)
    return zipfile_name


def move_file(file_path, dest):
    """
    move given file to the specified destination
    """
    try:
        shutil.move(file_path, dest)
    except Exception as e:
        print_fail("%s : ERROR: Backing up log files to %s failed. %s" % (get_timestamp(), dest, e))


def send_zipfile(zipfile_name, backup_folder_path):
    """
    upload the zip file to the endpoint
    """
    zipfile_path = os.path.join(get_logs_folder_path(), zipfile_name)
    # Get the endpoint url
    url = get_endpoint() + s.LOGS_ENDPOINT_PATH % (get_customer_id(), zipfile_name)
    is_post_success = False
    try:
        # post to Coolan endpoint
        is_post_success = post_file(url, get_username(), get_password(), zipfile_path)
        if is_post_success:
            print "%s : Zip file %s was sent to %s." % (get_timestamp(), zipfile_path, url)
    except Exception:
        print_fail(
            "%s : Failed sending zip file %s to Coolan endpoint %s." % (get_timestamp(), zipfile_path, url))

    # move the zip file to backup folder only if post successfully
    if is_post_success:
        move_file(zipfile_path, backup_folder_path)
        print "%s : Zip file %s was moved to the backup folder %s" % (get_timestamp(), zipfile_path, backup_folder_path)

    return is_post_success


def post_logs_to_endpoint():
    try:
        verify_root()
        if get_machine_id_setting() or get_boot_id_setting():
            print "%s : Started sending logs to Coolan endpoint." % get_timestamp()

            # first check if there are any zip files waiting to be sent
            for zip_file_path in glob.glob(get_logs_folder_path() + "/*.zip"):
                send_zipfile(os.path.basename(zip_file_path), get_backup_folder_path())

            # then check if there are any unzipped logs waiting to be sent
            zip_and_send_log_files(time_series=True)
            zip_and_send_log_files(time_series=False)
            if is_config_on('send_custom_logs_to_endpoint'):
                zip_and_send_custom_log_files()

            print "%s : Finished sending logs to Coolan endpoint." % get_timestamp()
        else:
            pass
    except:
        print_fail("%s : Failed sending logs to Coolan endpoint." % get_timestamp())
        exc_type, exc_value, exc_traceback = sys.exc_info()
        traceback.print_exception(exc_type, exc_value, exc_traceback, limit=2, file=sys.stdout)
        os._exit(1)
    finally:
        sys.stdout.flush()


def zip_and_send_log_files(time_series=False):
    to_send_file_paths = []
    backup_folder_path = get_backup_folder_path()

    if time_series:
        log_file_pattern = "/" + s.TIME_SERIES + "-*.log"
    else:
        log_file_pattern = "/" + s.RAWDATA + "-*.log"

    for log_file_path in glob.glob(get_logs_folder_path() + log_file_pattern):
        to_send_file_paths.append(log_file_path)
        # set the limit of number of files to send at once
        if len(to_send_file_paths) == s.NUMBER_OF_FILES_TO_ZIP:
            if time_series:
                zipfile_name = generate_time_series_log_filename() + '.zip'
            else:
                zipfile_name = generate_log_filename() + '.zip'
            send_zipfile(zip_log_files(to_send_file_paths, zipfile_name), backup_folder_path)
            # Start new array of files to send
            to_send_file_paths = []
    if len(to_send_file_paths) > 0:
        if time_series:
            zipfile_name = generate_time_series_log_filename() + '.zip'
        else:
            zipfile_name = generate_log_filename() + '.zip'
        send_zipfile(zip_log_files(to_send_file_paths, zipfile_name), backup_folder_path)


def zip_and_send_custom_log_files():
    to_send_file_paths = []
    log_file_pattern = "/" + s.CUSTOM_DATA + "-*.log"
    custom_backup_folder_path = get_custom_backup_folder_path()

    for log_file_path in glob.glob(get_custom_logs_folder_path() + log_file_pattern):
        to_send_file_paths.append(log_file_path)
        # set the limit of number of files to send at once
        if len(to_send_file_paths) == s.NUMBER_OF_FILES_TO_ZIP:
            zipfile_name = generate_custom_log_filename() + '.zip'
            send_zipfile(zip_log_files(to_send_file_paths, zipfile_name), custom_backup_folder_path)
            # Start new array of files to send
            to_send_file_paths = []
    if len(to_send_file_paths) > 0:
        zipfile_name = generate_custom_log_filename() + '.zip'
        send_zipfile(zip_log_files(to_send_file_paths, zipfile_name), custom_backup_folder_path)


if __name__ == "__main__":
    post_logs_to_endpoint()
