#!/usr/bin/env python
# Coolan Confidential
# Copyright (c) Coolan, Inc 2014 All Rights Reserved
#
# NOTICE:  All information contained herein is, and remains
# the property of Coolan, Inc.  The intellectual and
# technical concepts contained #herein are proprietary to
# Coolan, Inc and may be covered by U.S. and Foreign Patents,
# patents in process, and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material
# is strictly forbidden unless prior written permission is obtained
# from Coolan, Inc.
# This code and information are provided "as is" without warranty of any
# kind, either expressed or implied, including but not limited to the
# implied warranties of mechantability and/or fitness for a
# particular purpose.

import signal

from common import *
import settings as s
from collector import run_commands


def run_no_support_time_series_commands(log_path, commands, interval):
    for cmd in set(commands):
        args = []
        args.extend(shlex.split(cmd))
        command_full_path = find_command(args[0])
        if command_full_path:
            # adding the full path of the command to the command we get from commands.properties
            args[0] = command_full_path
            run_command_thread = RunCommandInThread(args, log_path, interval)
            timeout = int(config.get(s.SETTINGS, 'time_series_commands_time_out_seconds'))
            run_command_thread.run(timeout=timeout)

        # this command does not exist on this machine
        else:
            commands.remove(cmd)
            err = "Command %s doesn't exist.\n" % args[0]
            write_timeseries_output_to_log("", err, args, log_path, interval, time.time())


def run_support_time_series_commands(commands, interval, count):
    begin_end_tag_map = {}
    for command in commands:
        log_file_path = os.path.join(get_logs_folder_path(), generate_time_series_individual_log_filename(command))
        args = []
        args.extend(shlex.split(command))
        command_full_path = find_command(args[0])
        if command_full_path:
            args[0] = command_full_path
            full_command = " ".join(args)
            shell_command = full_command + " " + str(interval) + " " + str(count) + " > " + log_file_path
            # The os.setsid() is passed in the argument preexec_fn so
            # it's run after the fork() and before  exec() to run the shell.
            proc = subprocess.Popen(shell_command, shell=True, preexec_fn=os.setsid)
            begin_end_tag_map[log_file_path] = {"pid": proc.pid, "proc": proc, "command": args}
        # this command does not exist on this machine
        else:
            log_file = open(log_file_path, 'a')
            err = "Command %s doesn't exist.\n" % args[0]
            log_file.write(err)
            begin_end_tag_map[log_file_path] = {"pid": "", "proc": None, "command": args}
    return begin_end_tag_map


def main(interval=1, count=1):
    try:
        verify_root()
        start_time = time.time()
        print "%s : Started collecting time series data at an interval of %s seconds %s times." % (get_timestamp(), interval, count)

        boot_id = get_boot_id_setting()
        machine_id = get_machine_id_setting()

        # Get all the time series commands that support interval and count arguments
        commands_set = get_commands(s.TIME_SERIES_COMMANDS)
        support_time_series_commands_start_time = time.time()
        begin_end_tag_map = run_support_time_series_commands(commands_set, interval, count)

        # Get all the time series commands that do not support interval and count arguments
        no_support_commands_set = get_commands(s.TIME_SERIES_NO_SUPPORT_COMMANDS)
        if is_config_on('run_vm_commands'):
            # Add VM commands
            no_support_commands_set = no_support_commands_set.union(get_commands(s.VM_TIME_SERIES_COMMANDS))
        log_path = "%s.log%s" % (os.path.join(get_logs_folder_path(), generate_time_series_log_filename()), s.TMP_FILE_EXTENSION)

        run_count = 0
        while run_count < int(count):
            run_count += 1
            loop_start_time = time.time()
            # create one thread for each time series command that does not have continuous query support
            run_no_support_time_series_commands(log_path, no_support_commands_set, interval)

            sleep_seconds = loop_start_time + float(interval) - time.time()
            if sleep_seconds < 0:
                sleep_seconds = 0
            time.sleep(float(sleep_seconds))

        # append all the individual time series logs to the main time series log
        for time_series_individual_file_path in begin_end_tag_map.keys():
            command_map = begin_end_tag_map.get(time_series_individual_file_path)
            if command_map:

                # Check if child process has terminated. Set and return return code attribute. Doing this to prevent zombie processes
                proc = command_map.get("proc")
                if proc:
                    proc.poll()

                pid = command_map.get("pid")
                if pid:
                    # if the command is still running, give it 5 more seconds to finish
                    if is_pid_running(pid):
                        timeout = int(config.get(s.SETTINGS, 'time_series_commands_time_out_seconds'))
                        time.sleep(timeout)
                    # there is something wrong with the command, kill the process
                    if is_pid_running(pid):
                        print "%s : Collecting data at %s is taking too long. Killing the process pid %s. " % (get_timestamp(), time_series_individual_file_path, pid)
                        os.killpg(pid, signal.SIGTERM)  # Send the signal to all the process groups

                time_series_individual_file = open(time_series_individual_file_path, 'r')
                if pid:
                    write_timeseries_output_to_log(time_series_individual_file.read(), "", command_map.get("command"), log_path, interval, support_time_series_commands_start_time)
                else:
                    write_timeseries_output_to_log("", time_series_individual_file.read(), command_map.get("command"), log_path, interval, support_time_series_commands_start_time)
                # remove the individual time series log after appending it to the main time series log
                os.remove(time_series_individual_file_path)

            else:
                os.remove(time_series_individual_file_path)
                print "%s : Error occurred when collecting time series data at %s . The file was removed. " % (get_timestamp(), time_series_individual_file_path)

        # run the common periodic commands
        common_periodic_commands_set = get_commands(s.COMMON_PERIODIC_COMMANDS)
        run_commands(common_periodic_commands_set, log_path)

        # Print the general information section
        log_file = open(log_path, 'a')
        log_file.write(s.BEGIN_TIMESERIES_MSG % (int(start_time), interval, "General Information"))
        log_file.write("collectorVersion : %s\n" % s.COLLECTOR_VERSION)
        log_file.write("pythonVersion : %s\n" % sys.version)
        log_file.write("timeSeriesCollectorStartTimeSeconds : %s\n" % int(start_time))
        end_time = time.time()
        log_file.write("timeSeriesCollectorEndTimeSeconds : %s\n" % int(end_time))
        log_file.write("timeSeriesCollectorRunningTimeSeconds : %s\n" % str(end_time - start_time))
        log_file.write("currentMachineId : %s\n" % machine_id)
        log_file.write("currentBootId : %s\n" % boot_id)
        print_dmidecode_info(log_file)
        log_file.write(s.END_TIMESERIES_MSG % (int(start_time), interval, "General Information"))

        log_file.close()

        # Rename the log file to remove the tmp file extension
        new_log_path = log_path.replace(s.TMP_FILE_EXTENSION, "")
        os.rename(log_path, new_log_path)
        print "%s : Finished collecting time series data at an interval of %s seconds %s times. %s has been created. " % (get_timestamp(), interval, count, new_log_path)
    except:
        print_fail("%s : There are errors when generating the time series logs." % get_timestamp())
        exc_type, exc_value, exc_traceback = sys.exc_info()
        traceback.print_exception(exc_type, exc_value, exc_traceback, limit=2, file=sys.stdout)
        os._exit(1)
    finally:
        sys.stdout.flush()

if __name__ == "__main__":
    main()
